/*
 * Copyright (c) 2020, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAP2str(a) #a

//#define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

// define if shows elapsed time
//#define SNAPLIB_SHOW_ELAPSED_TIME

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.1le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
void show_dlist2(double *list, unsigned int len, unsigned int cols) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
    if ( (i+1)%cols == 0 ) {
      printf("\n");
    }
  }
  printf("\n");
}
#endif // defined
#endif // SNAPLIB_DEBUG

/* ================== START [agcd_ssgcd] ================== */

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
int SNAP_agcd_uvgcd_iteration_9_2i_lsy(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho, const SNAPBASEREAL_t threshold, const unsigned int max_iterations);
int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t *sigma_m1, gmp_randstate_t rstate);
int SNAP_agcd_uvgcd_solving_10_2_lsy(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r);
#endif // defined

typedef enum { SNAP_SSGCDEX_REFINEMENT_NONE, SNAP_SSGCDEX_REFINEMENT_UVGCD,
               SNAP_SSGCDEX_COFACTORS_BY_UVGCD, SNAP_SSGCDEX_COFACTORS_BY_SVD,
               SNAP_SSGCDEX_FROBENIUS_NORMAL, SNAP_SSGCDEX_FROBENIUS_WEIGHTED } mode_type;
/*!
 *  @brief SSGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the SSGCD algorithm following Schost and Spaenlehauer
 *   in 2016. (see Found Comput. Math. 16:457-492).
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0. GCD by SVD (not UVGCD) if ((mode/10)%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_ssgcd_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  // local variables
  SNAPMTX_t *basisE, *ptB;
  SNAPMTX_t matSyl, matU, matVT, matSVT, matA, vecB, vec_tmp;
  SNAPBASE_t *ptE, *ptE2, *ptE3;
  SNAPMTXREAL_t vecS;
  SNAPBASEREAL_t *ptR;
  SNAPBASEREAL_t norm_f, norm_g, norm_c, relative_tolerance, frobenius_norm_f, inv_frobenius_norm_f, frobenius_norm_g, inv_frobenius_norm_g, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g, poly_dummy_f, poly_dummy_g;
  int result_code = 0;
  unsigned int i, j, jj, sizeP, sizeQ, sizeR, sizeD;
  unsigned int loop_condition, loop_count;
  mode_type mode_refinement, mode_cofactors;
#ifdef SNAP_SUPD_T
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // SNAP_SUPD_T
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // initialization
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_UVGCD;
  }
  if ( (mode_opt/10)%10 == 0 ) {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_UVGCD;
  } else {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_SVD;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  sizeD = SNAP_degree(poly_f)+SNAP_degree(poly_g)+2;
  sizeP = SNAP_degree(poly_f)+SNAP_degree(poly_g)-d+1;   // rows
  sizeQ = SNAP_degree(poly_f)+SNAP_degree(poly_g)-2*d+2; // cols
  if ( sizeP < sizeQ ) {
    sizeR = sizeP - 1;
  } else {
    sizeR = sizeQ - 1;
  }
  SNAPMTX_init2(matSyl, sizeP, sizeQ);
  SNAPMTX_init2(matU, sizeP, sizeP);
  SNAPMTX_init2(matVT, sizeQ, sizeQ);
  SNAPMTX_init2(matSVT, sizeQ, sizeQ);
  SNAPMTXREAL_init2(vecS, 1, sizeQ);
  SNAPMTX_init2(matA, (sizeP-sizeR)*(sizeQ-sizeR), sizeD);
  if ( sizeD < (sizeP-sizeR)*(sizeQ-sizeR) ) {
    SNAPMTX_init2(vecB, 1, (sizeP-sizeR)*(sizeQ-sizeR));
  } else {
    SNAPMTX_init2(vecB, 1, sizeD);
  }
  SNAPMTX_init2(vec_tmp, 1, sizeP);
#ifdef SNAP_SUPD_T
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*sizeD);
#endif // defined
  basisE = (SNAPMTX_t*) malloc(sizeof(SNAPMTX_t)*sizeD);
  for ( i=1, ptB=basisE; i<=sizeD; i++, ptB++ ) {
    SNAPMTX_init2(*ptB, sizeP, sizeQ);
  }
  SNAP_init(poly_dummy_f);
  SNAP_init(poly_dummy_g);
  SNAPBASEREAL_init(frobenius_norm_f);
  SNAPBASEREAL_init(inv_frobenius_norm_f);
  SNAPBASEREAL_init(frobenius_norm_g);
  SNAPBASEREAL_init(inv_frobenius_norm_g);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(norm_c);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
  /* ====================== orthonormal basis E ====================== */
  // coefficient matrices of each coefficient of polynomials.
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 1.0);
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 0.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 1.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 0.0);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_f, SNAP_degree(poly_g)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_f, frobenius_norm_f);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_g, SNAP_degree(poly_f)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_g, frobenius_norm_g);
  for ( i=0, ptB=basisE; i<=SNAP_degree(poly_f); i++, ptB++ ) {
    SNAP_set_coef_basereal(poly_dummy_f, SNAP_degree(poly_f)-i, inv_frobenius_norm_f);
    SNAP_degree(poly_dummy_f) = SNAP_degree(poly_f);
    SNAP_degree(poly_dummy_g) = SNAP_degree(poly_g);
    SNAPMTX_subresultant_mapping_matrix_dsc(*ptB, poly_dummy_f, poly_dummy_g, d-1);
    SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f)-i, 0.0);
  }
  for ( i=0; i<=SNAP_degree(poly_g); i++, ptB++ ) {
    SNAP_set_coef_basereal(poly_dummy_g, SNAP_degree(poly_g)-i, inv_frobenius_norm_g);
    SNAP_degree(poly_dummy_f) = SNAP_degree(poly_f);
    SNAP_degree(poly_dummy_g) = SNAP_degree(poly_g);
    SNAPMTX_subresultant_mapping_matrix_dsc(*ptB, poly_dummy_f, poly_dummy_g, d-1);
    SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g)-i, 0.0);
  }
  /* ====================== preparation for the main loop ====================== */
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (init)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
  elapsed_time = snap_getrusageinsec();
  double elapsed_time_for_Syl=0, elapsed_time_for_SVD=0, elapsed_time_for_MT=0, elapsed_time_for_AB=0, elapsed_time_for_others=0;
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < max_iterations ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Sylvester resultant matrix
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Singular value decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'A', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTXREAL_elems(vecS), SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
      printf("ERROR in SNAP_agcd_ssgcd_ex: dgesdd could not compute the SVD.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_SVD += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#ifdef SNAPLIB_DEBUG
printf("matU\n");
show_sdmd(matU);
printf("matVT\n");
show_sdmd(matVT);
#endif // SNAPLIB_DEBUG
    // construct non-structured low-rank approximation (overrides matSyl!!!)
    // Sylvester resultant matrix since dgesdd destroyed matSyl!!
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    for ( i=0, ptE=SNAPMTX_elems(matVT), ptE2=SNAPMTX_elems(matSVT), ptR=SNAPMTX_elems(vecS); i < sizeR; i++, ptE+=SNAPMTX_csize(matVT), ptE2+=SNAPMTX_csize(matSVT), ptR++ ) {
#ifdef SNAP_SUPD_T
      cblas_dcopy(SNAPMTX_csize(matVT), ptE, 1, ptE2, 1);
      cblas_dscal(SNAPMTX_csize(matVT), *ptR, ptE2, 1);
#else
      IMPLEMENT HERE
#endif
    }
#ifdef SNAP_SUPD_T
    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), sizeR, 1.0, SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matSVT), SNAPMTX_csize(matSVT), -1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl));
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_DEBUG
printf("matMT\n");
show_sdmd(matSyl);
#endif // SNAPLIB_DEBUG
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_MT += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // preparation for correction terms -- matA and vecB
    for ( i=0, ptE=SNAPMTX_elems(matA), ptE2=SNAPMTX_elems(vecB); i < sizeP-sizeR; i++ ) {
      for ( j=0; j < sizeQ-sizeR; j++ ) {
        for ( jj=0, ptB=basisE; jj < sizeD; jj++, ptE++, ptB++ ) {
#ifdef SNAP_SUPD_T
          cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(*ptB), SNAPMTX_cols(*ptB), 1.0, SNAPMTX_elems(*ptB), SNAPMTX_csize(*ptB), SNAPMTX_elemp(matU, 0, sizeR+i), SNAPMTX_csize(matU), 0.0, SNAPMTX_elems(vec_tmp), 1);
#else
          IMPLEMENT HERE
#endif
          SNAPMTX_BLAS_dot(*ptE, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
        }
      }
#ifdef SNAP_SUPD_T
      cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), 1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), SNAPMTX_elemp(matU, 0, sizeR+i), SNAPMTX_csize(matU), 0.0, SNAPMTX_elems(vec_tmp), 1);
#else
      IMPLEMENT HERE
#endif
      for ( j=0; j < sizeQ-sizeR; j++, ptE2++ ) {
        SNAPMTX_BLAS_dot(*ptE2, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
      }
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_AB += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#ifdef SNAPLIB_DEBUG
printf("matA\n");
show_sdmd(matA);
printf("vecB\n");
show_sdmd(vecB);
#endif // SNAPLIB_DEBUG
    // correction terms
#ifdef SNAP_SUPD_T
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matA); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matA), SNAPMTX_cols(matA), 1, SNAPMTX_elems(matA), SNAPMTX_csize(matA), SNAPMTX_elems(vecB), 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ssgcd_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_DEBUG
printf("correction term\n");
show_sdmd(vecB);
#endif // SNAPLIB_DEBUG
    // loop condition check
    SNAPMTXREAL_BLAS_nrm2(norm_c, sizeD, SNAPMTX_elems(vecB), 1);
#ifdef SNAPLIB_DEBUG
printf("correction term: norm = %e\n", norm_c);
#endif // SNAPLIB_DEBUG
    if ( SNAPBASEREAL_cmp(norm_c, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // update
    for ( i=0, ptE=SNAP_coefs(poly_f), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)); i<=SNAP_degree(poly_f); i++, ptE++, ptE2-- ) {
      SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_f);
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
    for ( i=0, ptE=SNAP_coefs(poly_g), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)+SNAP_degree(poly_g)+1); i<=SNAP_degree(poly_g); i++, ptE++, ptE2-- ) {
      SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_g);
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_others += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in SNAP_agcd_ssgcd_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in SNAP_agcd_ssgcd_ex: converged within %d iterations.\n", loop_count);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (Syl in loop)   : %.18e\n", elapsed_time_for_Syl);
  printf("elapsed_time (SVD in loop)   : %.18e\n", elapsed_time_for_SVD);
  printf("elapsed_time (MT in loop)    : %.18e\n", elapsed_time_for_MT);
  printf("elapsed_time (AB in loop)    : %.18e\n", elapsed_time_for_AB);
  printf("elapsed_time (others in loop): %.18e\n", elapsed_time_for_others);
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // now we have polt_f and poly_g (perturbed polynomials)
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
    SNAPBASE_set_d(*ptE, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  for ( i=0, ptE=tau, ptE2=zvect, ptE3=yvect; i < SNAPMTX_csize(matSyl); i++,ptE++,ptE2++,ptE3++ ) {
    SNAPBASE_init(*ptE);
    SNAPBASE_init(*ptE2);
    SNAPBASE_init(*ptE3);
  }
  if ( mode_cofactors == SNAP_SSGCDEX_COFACTORS_BY_SVD ) {
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTX_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
       result_code = -1;
       printf("ERROR in SNAP_agcd_ssgcd3_ex: dgesdd could not compute the SVD.\n");
       goto post_uvgcd;
    }
#endif // SNAP_SUPD_T
    SNAP_set_rbase(poly_g1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,0), SNAP_degree(poly_ing)-d+1);
    SNAP_set_rbase(poly_f1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf)-d+1);
    SNAP_self_neg(poly_f1);
  } else {
    // QR decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), tau) != 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd_ex: dgeqrf could not compute the QR decomposition.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // After this, zvect is the singular vector
#ifdef SNAP_SUPD_T
    if ( SNAP_agcd_uvgcd_iteration_7_3(matSyl, zvect, yvect, &sigma_m1, rstate) < 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
    SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
    SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
    SNAP_self_neg(poly_f1);
  }
  if ( SNAP_agcd_uvgcd_solving_10_2_lsy(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
      if (
#ifdef SNAP_SUPD_T
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#else
        IMPLEMENT HERE
#endif
      ) {
        result_code = -1;
      }
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSyl);
  SNAPMTX_clear(matU);
  SNAPMTX_clear(matSVT);
  SNAPMTX_clear(matVT);
  SNAPMTXREAL_clear(vecS);
  SNAPMTX_clear(matA);
  SNAPMTX_clear(vecB);
  SNAPMTX_clear(vec_tmp);
  for ( i=1, ptB=basisE; i<=sizeD; i++, ptB++ ) {
    SNAPMTX_clear(*ptB);
  }
  SNAP_clear(poly_dummy_f);
  SNAP_clear(poly_dummy_g);
  SNAPBASEREAL_clear(frobenius_norm_f);
  SNAPBASEREAL_clear(inv_frobenius_norm_f);
  SNAPBASEREAL_clear(frobenius_norm_g);
  SNAPBASEREAL_clear(inv_frobenius_norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_c);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
#ifdef SNAP_SUPD_T
  free(jpvt);
#endif // defined
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (post)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  return result_code;
}

/*!
 *  @brief SSGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the SSGCD algorithm following Schost and Spaenlehauer
 *   in 2016. (see Found Comput. Math. 16:457-492).
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning This NewtonSLRA/1 is customized for subresultant matrix and not use basis matrices.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0. GCD by SVD (not UVGCD) if ((mode/10)%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_ssgcd2_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  // local variables
  SNAPMTX_t matSyl, matU, matVT, matSVT, matA, vecB, vec_tmp;
  SNAPBASE_t *ptE, *ptE2, *ptE3;
  SNAPMTXREAL_t vecS;
  SNAPBASEREAL_t *ptR;
  SNAPBASEREAL_t norm_f, norm_g, norm_c, relative_tolerance, frobenius_norm_f, inv_frobenius_norm_f, frobenius_norm_g, inv_frobenius_norm_g, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g, poly_dummy_f, poly_dummy_g;
  int result_code = 0;
  unsigned int i, j, jj, sizeP, sizeQ, sizeR, sizeD;
  unsigned int loop_condition, loop_count;
  mode_type mode_refinement, mode_cofactors;
#ifdef SNAP_SUPD_T
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // SNAP_SUPD_T
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // initialization
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_UVGCD;
  }
  if ( (mode_opt/10)%10 == 0 ) {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_UVGCD;
  } else {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_SVD;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  sizeD = SNAP_degree(poly_f)+SNAP_degree(poly_g)+2;
  sizeP = SNAP_degree(poly_f)+SNAP_degree(poly_g)-d+1;   // rows
  sizeQ = SNAP_degree(poly_f)+SNAP_degree(poly_g)-2*d+2; // cols
  if ( sizeP < sizeQ ) {
    sizeR = sizeP - 1;
  } else {
    sizeR = sizeQ - 1;
  }
  SNAPMTX_init2(matSyl, sizeP, sizeQ);
  SNAPMTX_init2(matU, sizeP, sizeP);
  SNAPMTX_init2(matVT, sizeQ, sizeQ);
  SNAPMTX_init2(matSVT, sizeQ, sizeQ);
  SNAPMTXREAL_init2(vecS, 1, sizeQ);
  SNAPMTX_init2(matA, (sizeP-sizeR)*(sizeQ-sizeR), sizeD);
  if ( sizeD < (sizeP-sizeR)*(sizeQ-sizeR) ) {
    SNAPMTX_init2(vecB, 1, (sizeP-sizeR)*(sizeQ-sizeR));
  } else {
    SNAPMTX_init2(vecB, 1, sizeD);
  }
  SNAPMTX_init2(vec_tmp, 1, sizeP);
#ifdef SNAP_SUPD_T
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*sizeD);
#endif // defined
  SNAP_init(poly_dummy_f);
  SNAP_init(poly_dummy_g);
  SNAPBASEREAL_init(frobenius_norm_f);
  SNAPBASEREAL_init(inv_frobenius_norm_f);
  SNAPBASEREAL_init(frobenius_norm_g);
  SNAPBASEREAL_init(inv_frobenius_norm_g);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(norm_c);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
  /* ====================== orthonormal basis E ====================== */
  // coefficient matrices of each coefficient of polynomials.
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 1.0);
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 0.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 1.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 0.0);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_f, SNAP_degree(poly_g)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_f, frobenius_norm_f);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_g, SNAP_degree(poly_f)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_g, frobenius_norm_g);
  /* ====================== preparation for the main loop ====================== */
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (init)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
  elapsed_time = snap_getrusageinsec();
  double elapsed_time_for_Syl=0, elapsed_time_for_SVD=0, elapsed_time_for_MT=0, elapsed_time_for_AB=0, elapsed_time_for_others=0;
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < max_iterations ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Sylvester resultant matrix
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Singular value decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'A', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTXREAL_elems(vecS), SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
      printf("ERROR in SNAP_agcd_ssgcd2_ex: dgesdd could not compute the SVD.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_SVD += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#ifdef SNAPLIB_DEBUG
printf("matU\n");
show_sdmd(matU);
printf("matVT\n");
show_sdmd(matVT);
#endif // SNAPLIB_DEBUG
    // construct non-structured low-rank approximation (overrides matSyl!!!)
    // Sylvester resultant matrix since dgesdd destroyed matSyl!!
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    for ( i=0, ptE=SNAPMTX_elems(matVT), ptE2=SNAPMTX_elems(matSVT), ptR=SNAPMTX_elems(vecS); i < sizeR; i++, ptE+=SNAPMTX_csize(matVT), ptE2+=SNAPMTX_csize(matSVT), ptR++ ) {
#ifdef SNAP_SUPD_T
      cblas_dcopy(SNAPMTX_csize(matVT), ptE, 1, ptE2, 1);
      cblas_dscal(SNAPMTX_csize(matVT), *ptR, ptE2, 1);
#else
      IMPLEMENT HERE
#endif
    }
#ifdef SNAP_SUPD_T
    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), sizeR, 1.0, SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matSVT), SNAPMTX_csize(matSVT), -1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl));
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_DEBUG
printf("matMT\n");
show_sdmd(matSyl);
#endif // SNAPLIB_DEBUG
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_MT += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // preparation for correction terms -- matA and vecB
    for ( i=0, ptE=SNAPMTX_elems(matA), ptE2=SNAPMTX_elems(vecB); i < sizeP-sizeR; i++ ) {
      for ( j=0; j < sizeQ-sizeR; j++ ) {
        for ( jj=0; jj < sizeD; jj++, ptE++ ) {
          SNAPBASE_t *ptU, *ptT;
          unsigned int k;
          if ( jj <= SNAP_degree(poly_f) ) {
            for ( k=0, ptU=SNAPMTX_elemp(matU, jj, sizeR+i),ptT=SNAPMTX_elems(vec_tmp); k < SNAP_degree(poly_g)-d+1; k++,ptU+=SNAPMTX_csize(matU),ptT++ ) {
                SNAPBASE_mul(*ptT, *ptU, inv_frobenius_norm_f);
            }
            for ( ; k < SNAPMTX_cols(matVT); k++,ptT++ ) {
                SNAPBASE_set_ui(*ptT, 0);
            }
          } else {
            for ( k=0,ptT=SNAPMTX_elems(vec_tmp); k < SNAP_degree(poly_g)-d+1; k++,ptT++ ) {
                SNAPBASE_set_ui(*ptT, 0);
            }
            for ( ptU=SNAPMTX_elemp(matU, jj-SNAP_degree(poly_f)-1, sizeR+i) ; k < SNAPMTX_cols(matVT); k++,ptU+=SNAPMTX_csize(matU),ptT++ ) {
                SNAPBASE_mul(*ptT, *ptU, inv_frobenius_norm_g);
            }
          }
          SNAPMTX_BLAS_dot(*ptE, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
        }
      }
#ifdef SNAP_SUPD_T
      cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), 1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), SNAPMTX_elemp(matU, 0, sizeR+i), SNAPMTX_csize(matU), 0.0, SNAPMTX_elems(vec_tmp), 1);
#else
      IMPLEMENT HERE
#endif
      for ( j=0; j < sizeQ-sizeR; j++, ptE2++ ) {
        SNAPMTX_BLAS_dot(*ptE2, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
      }
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_AB += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // correction terms
#ifdef SNAP_SUPD_T
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matA); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matA), SNAPMTX_cols(matA), 1, SNAPMTX_elems(matA), SNAPMTX_csize(matA), SNAPMTX_elems(vecB), 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ssgcd2_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
    // loop condition check
    SNAPMTXREAL_BLAS_nrm2(norm_c, sizeD, SNAPMTX_elems(vecB), 1);
    if ( SNAPBASEREAL_cmp(norm_c, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // update
    for ( i=0, ptE=SNAP_coefs(poly_f), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)); i<=SNAP_degree(poly_f); i++, ptE++, ptE2-- ) {
      SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_f);
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
    for ( i=0, ptE=SNAP_coefs(poly_g), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)+SNAP_degree(poly_g)+1); i<=SNAP_degree(poly_g); i++, ptE++, ptE2-- ) {
      SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_g);
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_others += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in SNAP_agcd_ssgcd2_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in SNAP_agcd_ssgcd2_ex: converged within %d iterations.\n", loop_count);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (Syl in loop)   : %.18e\n", elapsed_time_for_Syl);
  printf("elapsed_time (SVD in loop)   : %.18e\n", elapsed_time_for_SVD);
  printf("elapsed_time (MT in loop)    : %.18e\n", elapsed_time_for_MT);
  printf("elapsed_time (AB in loop)    : %.18e\n", elapsed_time_for_AB);
  printf("elapsed_time (others in loop): %.18e\n", elapsed_time_for_others);
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // now we have polt_f and poly_g (perturbed polynomials)
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
    SNAPBASE_set_d(*ptE, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  for ( i=0, ptE=tau, ptE2=zvect, ptE3=yvect; i < SNAPMTX_csize(matSyl); i++,ptE++,ptE2++,ptE3++ ) {
    SNAPBASE_init(*ptE);
    SNAPBASE_init(*ptE2);
    SNAPBASE_init(*ptE3);
  }
  if ( mode_cofactors == SNAP_SSGCDEX_COFACTORS_BY_SVD ) {
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTX_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
       result_code = -1;
       printf("ERROR in SNAP_agcd_ssgcd3_ex: dgesdd could not compute the SVD.\n");
       goto post_uvgcd;
    }
#endif // SNAP_SUPD_T
    SNAP_set_rbase(poly_g1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,0), SNAP_degree(poly_ing)-d+1);
    SNAP_set_rbase(poly_f1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf)-d+1);
    SNAP_self_neg(poly_f1);
  } else {
    // QR decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), tau) != 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd2_ex: dgeqrf could not compute the QR decomposition.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // After this, zvect is the singular vector
#ifdef SNAP_SUPD_T
    if ( SNAP_agcd_uvgcd_iteration_7_3(matSyl, zvect, yvect, &sigma_m1, rstate) < 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd2_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
    SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
    SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
    SNAP_self_neg(poly_f1);
  }
  if ( SNAP_agcd_uvgcd_solving_10_2_lsy(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
/*
      if (
#ifdef SNAP_SUPD_T
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#else
        IMPLEMENT HERE
#endif
      ) {
        result_code = -1;
      }
*/
#ifdef SNAP_SUPD_T
      SNAP_agcd_uvgcd_iteration_9_2i_lsy(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho, threshold, max_iterations);
#else
        IMPLEMENT HERE
#endif
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSyl);
  SNAPMTX_clear(matU);
  SNAPMTX_clear(matSVT);
  SNAPMTX_clear(matVT);
  SNAPMTXREAL_clear(vecS);
  SNAPMTX_clear(matA);
  SNAPMTX_clear(vecB);
  SNAPMTX_clear(vec_tmp);
  SNAP_clear(poly_dummy_f);
  SNAP_clear(poly_dummy_g);
  SNAPBASEREAL_clear(frobenius_norm_f);
  SNAPBASEREAL_clear(inv_frobenius_norm_f);
  SNAPBASEREAL_clear(frobenius_norm_g);
  SNAPBASEREAL_clear(inv_frobenius_norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_c);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
#ifdef SNAP_SUPD_T
  free(jpvt);
#endif // defined
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (post)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  return result_code;
}

// enable only when implementing.
#define SNAP_SUPD_T

/*!
 *  @brief SSGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of our relaxed SSGCD algorithm that is an improved version of
 *           NewtonSLRA/1 algorithm by Schost and Spaenlehauer in 2016. (see Found Comput. Math. 16:457-492).
 *           In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning This NewtonSLRA/1 is customized for subresultant matrix and not use basis matrices.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold_relaxed The threshold for iterations by relaxed NewtonSLRA/1
               (relaxed NewtonSLRA/1 is used while the smallest singular value is larger than or equal to this value).
 *  @param[in] deficient The deficient dimension of left null space in the relaxed NewtonSLRA/1.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0. GCD by SVD (not UVGCD) if ((mode/10)%10) is not 0.
               The weighted Frobenius norm is used if ((mode/100)%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_ssgcd3_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold_relaxed, const unsigned int deficient, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  // local variables
  SNAPMTX_t matSyl, matU, matVT, matSVT, matA, vecB, vec_tmp;
  SNAPBASE_t *ptE, *ptE2, *ptE3, *ptU, *ptT;
  SNAPMTXREAL_t vecS;
  SNAPBASEREAL_t *ptR;
  SNAPBASEREAL_t norm_f, norm_g, norm_c, relative_tolerance, frobenius_norm_f, inv_frobenius_norm_f, frobenius_norm_g, inv_frobenius_norm_g, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g, poly_dummy_f, poly_dummy_g;
  int result_code = 0;
  unsigned int i, j, jj, k, sizeP, sizeQ, sizeR, sizeD, relaxed = 0;
  unsigned int loop_condition, loop_count;
  mode_type mode_refinement, mode_cofactors, mode_frobenius;
#ifdef SNAP_SUPD_T
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // SNAP_SUPD_T
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // initialization
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_SSGCDEX_REFINEMENT_UVGCD;
  }
  if ( (mode_opt/10)%10 == 0 ) {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_UVGCD;
  } else {
    mode_cofactors = SNAP_SSGCDEX_COFACTORS_BY_SVD;
  }
  if ( (mode_opt/100)%10 == 0 ) {
    mode_frobenius = SNAP_SSGCDEX_FROBENIUS_NORMAL;
  } else {
    mode_frobenius = SNAP_SSGCDEX_FROBENIUS_WEIGHTED;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  sizeD = SNAP_degree(poly_f)+SNAP_degree(poly_g)+2;
  sizeP = SNAP_degree(poly_f)+SNAP_degree(poly_g)-d+1;   // rows
  sizeQ = SNAP_degree(poly_f)+SNAP_degree(poly_g)-2*d+2; // cols
  if ( sizeP < sizeQ ) {
    sizeR = sizeP - 1;
  } else {
    sizeR = sizeQ - 1;
  }
  SNAPMTX_init2(matSyl, sizeP, sizeQ);
  SNAPMTX_init2(matU, sizeP, sizeP);
  SNAPMTX_init2(matVT, sizeQ, sizeQ);
  SNAPMTX_init2(matSVT, sizeQ, sizeQ);
  SNAPMTXREAL_init2(vecS, 1, sizeQ);
  SNAPMTX_init2(matA, (sizeP-sizeR)*(sizeQ-sizeR), sizeD);
  if ( sizeD < (sizeP-sizeR)*(sizeQ-sizeR) ) {
    SNAPMTX_init2(vecB, 1, (sizeP-sizeR)*(sizeQ-sizeR));
  } else {
    SNAPMTX_init2(vecB, 1, sizeD);
  }
  SNAPMTX_init2(vec_tmp, 1, sizeP);
#ifdef SNAP_SUPD_T
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*sizeD);
#endif // defined
  SNAP_init(poly_dummy_f);
  SNAP_init(poly_dummy_g);
  SNAPBASEREAL_init(frobenius_norm_f);
  SNAPBASEREAL_init(inv_frobenius_norm_f);
  SNAPBASEREAL_init(frobenius_norm_g);
  SNAPBASEREAL_init(inv_frobenius_norm_g);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(norm_c);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
  /* ====================== orthonormal basis E ====================== */
  // coefficient matrices of each coefficient of polynomials.
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 1.0);
  SNAP_set_coef_d(poly_dummy_f, SNAP_degree(poly_f), 0.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 1.0);
  SNAP_set_coef_d(poly_dummy_g, SNAP_degree(poly_g), 0.0);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_f, SNAP_degree(poly_g)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_f, frobenius_norm_f);
  SNAPBASEREAL_sqrt_ui(frobenius_norm_g, SNAP_degree(poly_f)-d+1);
  SNAPBASEREAL_rec(inv_frobenius_norm_g, frobenius_norm_g);
  /* ====================== preparation for the main loop ====================== */
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (init)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
  elapsed_time = snap_getrusageinsec();
  double elapsed_time_for_Syl=0, elapsed_time_for_SVD=0, elapsed_time_for_MT=0, elapsed_time_for_AB=0, elapsed_time_for_others=0;
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < max_iterations ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Sylvester resultant matrix
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
    if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
      for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_rows(matSyl); i++ ) {
        for ( j=0; j < SNAP_degree(poly_g)-d+1; j++, ptE++ ) {
          SNAPBASE_mul(*ptE, *ptE, inv_frobenius_norm_f);
        }
        for ( ; j < SNAPMTX_cols(matSyl); j++, ptE++ ) {
          SNAPBASE_mul(*ptE, *ptE, inv_frobenius_norm_g);
        }
      }
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // Singular value decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'A', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTXREAL_elems(vecS), SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
      printf("ERROR in SNAP_agcd_ssgcd3_ex: dgesdd could not compute the SVD.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_SVD += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#ifdef SNAPLIB_DEBUG
printf("matU\n");
show_sdmd(matU);
printf("matVT\n");
show_sdmd(matVT);
#endif // SNAPLIB_DEBUG
    // the deficient dimension of the left null space in the relaxed NewtonSLRA/1
    if ( d <= 1 || SNAPBASEREAL_cmp(SNAPMTXREAL_elem(vecS,0,sizeR), threshold_relaxed) <= 0 ) {
      relaxed = 0;
    } else if ( sizeP < sizeR+deficient+1 ) {
      relaxed = sizeP-sizeR-1;
    } else {
      relaxed = deficient;
    }
#ifdef SNAPLIB_DEBUG
    printf("INFO: the dimension deficient: %d (sigma=%e, thres=%e)\n", relaxed, SNAPMTXREAL_elem(vecS,0,sizeR), threshold_relaxed);
#endif // SNAPLIB_DEBUG
    // construct non-structured low-rank approximation (overrides matSyl!!!)
    // Sylvester resultant matrix since dgesdd destroyed matSyl!!
    for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
      SNAPBASE_set_d(*ptE, 0.0);
    }
    SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
    if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
      for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_rows(matSyl); i++ ) {
        for ( j=0; j < SNAP_degree(poly_g)-d+1; j++, ptE++ ) {
          SNAPBASE_mul(*ptE, *ptE, inv_frobenius_norm_f);
        }
        for ( ; j < SNAPMTX_cols(matSyl); j++, ptE++ ) {
          SNAPBASE_mul(*ptE, *ptE, inv_frobenius_norm_g);
        }
      }
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_Syl += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    for ( i=0, ptE=SNAPMTX_elems(matVT), ptE2=SNAPMTX_elems(matSVT), ptR=SNAPMTX_elems(vecS); i < sizeR; i++, ptE+=SNAPMTX_csize(matVT), ptE2+=SNAPMTX_csize(matSVT), ptR++ ) {
#ifdef SNAP_SUPD_T
      cblas_dcopy(SNAPMTX_csize(matVT), ptE, 1, ptE2, 1);
      cblas_dscal(SNAPMTX_csize(matVT), *ptR, ptE2, 1);
#else
      IMPLEMENT HERE
#endif
      if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
        for ( k=0, ptT=ptE2; k < SNAP_degree(poly_g)-d+1; k++, ptT++ ) {
          SNAPBASE_mul(*ptT, *ptT, frobenius_norm_f);
        }
        for ( ; k < SNAPMTX_cols(matVT); k++, ptT++ ) {
          SNAPBASE_mul(*ptT, *ptT, frobenius_norm_g);
        }
      }
    }
#ifdef SNAP_SUPD_T
    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), sizeR, 1.0, SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matSVT), SNAPMTX_csize(matSVT), -1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl));
#else
    IMPLEMENT HERE
#endif
#ifdef SNAPLIB_DEBUG
printf("matMT\n");
show_sdmd(matSyl);
#endif // SNAPLIB_DEBUG
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_MT += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // preparation for correction terms -- matA and vecB
    for ( i=0, ptE=SNAPMTX_elems(matA), ptE2=SNAPMTX_elems(vecB); i < sizeP-sizeR-relaxed; i++ ) {
      for ( j=0; j < sizeQ-sizeR; j++ ) {
        for ( jj=0; jj < sizeD; jj++, ptE++ ) {
          if ( jj <= SNAP_degree(poly_f) ) {
            for ( k=0, ptU=SNAPMTX_elemp(matU, jj, sizeR+i),ptT=SNAPMTX_elems(vec_tmp); k < SNAP_degree(poly_g)-d+1; k++,ptU+=SNAPMTX_csize(matU),ptT++ ) {
                SNAPBASE_mul(*ptT, *ptU, inv_frobenius_norm_f);
                if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
                  SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_f);
                  // SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_f); // cancelled by W^{-1} for V^T.
                }
            }
            for ( ; k < SNAPMTX_cols(matVT); k++,ptT++ ) {
                SNAPBASE_set_ui(*ptT, 0);
            }
          } else {
            for ( k=0,ptT=SNAPMTX_elems(vec_tmp); k < SNAP_degree(poly_g)-d+1; k++,ptT++ ) {
                SNAPBASE_set_ui(*ptT, 0);
            }
            for ( ptU=SNAPMTX_elemp(matU, jj-SNAP_degree(poly_f)-1, sizeR+i) ; k < SNAPMTX_cols(matVT); k++,ptU+=SNAPMTX_csize(matU),ptT++ ) {
                SNAPBASE_mul(*ptT, *ptU, inv_frobenius_norm_g);
                if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
                  SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_g);
                  // SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_g); // cancelled by W^{-1} for V^T.
                }
            }
          }
          SNAPMTX_BLAS_dot(*ptE, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
        }
      }
#ifdef SNAP_SUPD_T
      cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), 1.0, SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), SNAPMTX_elemp(matU, 0, sizeR+i), SNAPMTX_csize(matU), 0.0, SNAPMTX_elems(vec_tmp), 1);
#else
      IMPLEMENT HERE
#endif
      if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_WEIGHTED ) {
        for ( k=0, ptT=SNAPMTX_elems(vec_tmp); k < SNAP_degree(poly_g)-d+1; k++, ptT++ ) {
          SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_f);
          // SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_f); // cancelled by W^{-1} for V^T.
        }
        for ( ; k < SNAPMTX_cols(matVT); k++, ptT++ ) {
          SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_g);
          // SNAPBASE_mul(*ptT, *ptT, inv_frobenius_norm_g); // cancelled by W^{-1} for V^T.
        }
      }
      for ( j=0; j < sizeQ-sizeR; j++, ptE2++ ) {
        SNAPMTX_BLAS_dot(*ptE2, SNAPMTX_cols(matVT), SNAPMTX_elemp(matVT, sizeR+j, 0), 1, SNAPMTX_elems(vec_tmp), 1);
      }
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_AB += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // correction terms
#ifdef SNAP_SUPD_T
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matA); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matA)-relaxed, SNAPMTX_cols(matA), 1, SNAPMTX_elems(matA), SNAPMTX_csize(matA), SNAPMTX_elems(vecB), 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ssgcd3_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
      result_code = -1;
      goto post;
    }
#else
    IMPLEMENT HERE
#endif
    // loop condition check
    SNAPMTXREAL_BLAS_nrm2(norm_c, sizeD, SNAPMTX_elems(vecB), 1);
    if ( SNAPBASEREAL_cmp(norm_c, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // update
    for ( i=0, ptE=SNAP_coefs(poly_f), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)); i<=SNAP_degree(poly_f); i++, ptE++, ptE2-- ) {
      if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_NORMAL ) {
        SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_f);
      }
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
    for ( i=0, ptE=SNAP_coefs(poly_g), ptE2=SNAPMTX_elemp(vecB, 0, SNAP_degree(poly_f)+SNAP_degree(poly_g)+1); i<=SNAP_degree(poly_g); i++, ptE++, ptE2-- ) {
      if ( mode_frobenius == SNAP_SSGCDEX_FROBENIUS_NORMAL ) {
        SNAPBASE_div_basereal(*ptE2, *ptE2, frobenius_norm_g);
      }
      SNAPBASE_add(*ptE, *ptE, *ptE2);
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time_for_others += snap_getrusageinsec()-elapsed_time;
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in SNAP_agcd_ssgcd3_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in SNAP_agcd_ssgcd3_ex: converged within %d iterations.\n", loop_count);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (Syl in loop)   : %.18e\n", elapsed_time_for_Syl);
  printf("elapsed_time (SVD in loop)   : %.18e\n", elapsed_time_for_SVD);
  printf("elapsed_time (MT in loop)    : %.18e\n", elapsed_time_for_MT);
  printf("elapsed_time (AB in loop)    : %.18e\n", elapsed_time_for_AB);
  printf("elapsed_time (others in loop): %.18e\n", elapsed_time_for_others);
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // now we have polt_f and poly_g (perturbed polynomials)
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, ptE=SNAPMTX_elems(matSyl); i < SNAPMTX_csize(matSyl)*SNAPMTX_rsize(matSyl); i++,ptE++ ) {
    SNAPBASE_set_d(*ptE, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSyl, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSyl));
  for ( i=0, ptE=tau, ptE2=zvect, ptE3=yvect; i < SNAPMTX_csize(matSyl); i++,ptE++,ptE2++,ptE3++ ) {
    SNAPBASE_init(*ptE);
    SNAPBASE_init(*ptE2);
    SNAPBASE_init(*ptE3);
  }
  if ( mode_cofactors == SNAP_SSGCDEX_COFACTORS_BY_SVD ) {
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl),
                         SNAPMTX_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
       result_code = -1;
       printf("ERROR in SNAP_agcd_ssgcd3_ex: dgesdd could not compute the SVD.\n");
       goto post_uvgcd;
    }
#endif // SNAP_SUPD_T
    SNAP_set_rbase(poly_g1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,0), SNAP_degree(poly_ing)-d+1);
    SNAP_set_rbase(poly_f1, SNAPMTX_elemp(matVT,SNAPMTX_cols(matSyl)-1,SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf)-d+1);
    SNAP_self_neg(poly_f1);
  } else {
    // QR decomposition
#ifdef SNAP_SUPD_T
    if ( LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSyl), SNAPMTX_cols(matSyl), SNAPMTX_elems(matSyl), SNAPMTX_csize(matSyl), tau) != 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd3_ex: dgeqrf could not compute the QR decomposition.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // After this, zvect is the singular vector
#ifdef SNAP_SUPD_T
    if ( SNAP_agcd_uvgcd_iteration_7_3(matSyl, zvect, yvect, &sigma_m1, rstate) < 0 ) {
      result_code = -1;
      printf("ERROR in SNAP_agcd_ssgcd3_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      goto post_uvgcd;
    }
#else
  IMPLEMENT HERE
#endif
    // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
    SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
    SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
    SNAP_self_neg(poly_f1);
  }
  if ( SNAP_agcd_uvgcd_solving_10_2_lsy(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
/*
      if (
#ifdef SNAP_SUPD_T
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#else
        IMPLEMENT HERE
#endif
      ) {
        result_code = -1;
      }
*/
#ifdef SNAP_SUPD_T
      SNAP_agcd_uvgcd_iteration_9_2i_lsy(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho, threshold, max_iterations);
#else
        IMPLEMENT HERE
#endif
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_SSGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSyl);
  SNAPMTX_clear(matU);
  SNAPMTX_clear(matSVT);
  SNAPMTX_clear(matVT);
  SNAPMTXREAL_clear(vecS);
  SNAPMTX_clear(matA);
  SNAPMTX_clear(vecB);
  SNAPMTX_clear(vec_tmp);
  SNAP_clear(poly_dummy_f);
  SNAP_clear(poly_dummy_g);
  SNAPBASEREAL_clear(frobenius_norm_f);
  SNAPBASEREAL_clear(inv_frobenius_norm_f);
  SNAPBASEREAL_clear(frobenius_norm_g);
  SNAPBASEREAL_clear(inv_frobenius_norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_c);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
#ifdef SNAP_SUPD_T
  free(jpvt);
#endif // defined
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (post)          : %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  return result_code;
}





